#!/usr/bin/perl

use strict;
use warnings;
use IPC::System::Simple qw(system capture);
use Cwd;
use Getopt::Long;
use List::Util qw(max min);
use Math::Trig;

my $help = 0;
my $strict_vrml = 0;

my $radius = 1;
my $height = 1;
my $twist = 1;
my $max_z = 4;
my $min_z = -4;
my @density = (512, 512, 64);
my $surface_density = 64;

$Getopt::Long::ignorecase = 0;
GetOptions("help" => \$help,
           "radius=f" => \$radius,
           "height=f" => \$height,
           "twist=f" => \$twist,
           "maxz=f" => \$max_z,
           "minz=f" => \$min_z,
           "surface-density=i" => \$surface_density) or die "Bad command line option!";
usage() if $help;

sub usage {
    printf "gen_hyperbolic_paraboloid options";
    printf "
\t--help\t\t\tprint this help message
\t--minz <minz>\t\tset the minimum z-coordinate <minz>
\t--maxz <maxz>\t\tset the maximum z-coordinate <maxz>
\t--surface-density <density>\t\tset the surface density to <density>
";
  exit(0);
}

my $v_min = -$twist;
my $v_max = $twist;

my $min_x = 0;
my $max_x = 1;

my $min_y = 0;
my $max_y = 1;

my @s1s = ( 1,  4, -4);
my @s1t = ( 1, -4,  4);
my @s2s = (-1, -4, -4);
my @s2t = (-1,  4,  4);

my @p0 = (-4,  1, -4);
my @p1 = ( 4,  1,  4);
my @p2 = ( 4, -1, -4);
my @p3 = (-4, -1,  4);

sub clip_t3 {
  my @res = ();
  my ($min_x, $max_x, $t3_min_x, $t3_max_x, $t3_infinity_x) = @_;
  if (($t3_min_x < $t3_infinity_x) && ($t3_infinity_x < $t3_max_x)) {
    # Exclude [$t3_min_x,$t3_max_x]
    if (($t3_min_x < $min_x) && ($max_x < $t3_max_x)) {
      return @res;
    }
    if (($t3_max_x < $min_x) || ($max_x < $t3_min_x)) {
      # print "1 min_x = $min_x, max_x = $max_x\n";
      push @res, $min_x, $max_x;
      return @res;
    }
    if ($min_x < $t3_min_x) {
      # print "2 min_x = $min_x, max_x = $max_x\n";
      push @res, $min_x, $t3_min_x;
    }
    if ($t3_max_x < $max_x) {
      # print "3 min_x = $min_x, max_x = $max_x\n";
      push @res, $t3_max_x, $max_x;
    }
    return @res;
  }
  # Include [$t3_min_x,$t3_max_x]
  $min_x = max($min_x, $t3_min_x);
  $max_x = min($t3_max_x, $max_x);
  # print "4 min_x = $min_x, max_x = $max_x\n";
  if ($min_x < $max_x) {
    push @res, $min_x, $max_x;
  }
  return @res;
}

sub get_curves {
  my $id = $_[0];
  my ($a, $b, $c, $d, $t3_0_x, $t3_1_x, $t3_infinity_x) = split(/ /, $_[1]);
  my $t3_min_x = min($t3_0_x, $t3_1_x);
  my $t3_max_x = max($t3_0_x, $t3_1_x);
  my @intervals = ();
  my @surfaces = ();

  if ($c == 0) {
    # A horizontal or diagonal line
    if ($d == 0) {
      # Illegal curve: Both c and d vanish!"
      return;
    }
    my ($y0, $y1);
    if ($a == 0) {
      # A horizontal line
      $y0 = $b / $d;
      $y1 = $y0;
      if (($min_y > $y1) || ($y0 > $max_y)) {
	# Empty curve
        return @surfaces;
      }
      push(@intervals,
	   clip_t3($min_x, $max_x, $t3_min_x, $t3_max_x, $t3_infinity_x));
    } else {
      # A diagonal line
      $a = $a / $d;
      $b = $b / $d;
      my $i0 = -$b / $a;
      my $i1 = (1 - $b) / $a;
      my $x0 = min($i0, $i1);
      my $x1 = max($i0, $i1);
      if (($x0 > 1) || ($x1 < 0)) {
	# Empty curve
        return @surfaces;
      }
      $x0 = $min_x if ($x0 < $min_x);
      $x1 = $max_x if ($max_x < $x1);
      $y0 = $a * $x0 + $b;
      $y1 = $a * $x1 + $b;
      push(@intervals,
	   clip_t3($x0, $x1, $t3_min_x, $t3_max_x, $t3_infinity_x));
    }
    for my $i (0 .. $#intervals/2-1) {
      my $x0 = $intervals[$i*2];
      my $x1 = $intervals[$i*2+1];
      # my $curve = sprintf("\\pcline(%f, %f)(%f, %f)", $x0, $y0, $x1, $y1);
    }
    return @surfaces;
  }

  if ($a * $d == $b * $c) {
    # A vertical line
    my $num = gcf($a, $b);
    my $den = gcf($c, $d);
    my $val = $num / $den;
    if (($min_y > $val) || ($val > $max_y)) {
      # Empty curve
      return @surfaces;
    }
    # my $curve = sprintf("\\pcline(%f, %f)(%f, %f)", $val, $min_y, $val, $max_y);
    return;
  }
  # Hyperbola
  my $y0_x = ($min_x * $d - $b) / ($a - $min_x * $c);
  my $y1_x = ($max_x * $d - $b) / ($a - $max_x * $c);
  my $y_min_x = min($y0_x, $y1_x);
  my $y_max_x = max($y0_x, $y1_x);
  my $asymptote_x = -$d / $c;
  my $asymptote_y = $a / $c;
  # print "y_min_x = $y_min_x\n";
  # print "y_max_x = $y_max_x\n";
  if (($min_y < $asymptote_y) && ($asymptote_y < $max_y)) {
    #  The hyperbola horizontal asymptote is in [min_y,max_y]
    if ($min_x < $y_min_x) {
      my $x1 = $y_min_x;
      $x1 = $max_x if ($max_x < $x1);
      push @intervals,
           clip_t3($min_x, $x1, $t3_min_x, $t3_max_x, $t3_infinity_x);
    }
    if ($y_max_x < $max_x) {
      my $x0 = $y_max_x;
      $x0 = $min_x if ($x0 < $min_x);
      push @intervals,
	   clip_t3($x0, $max_x, $t3_min_x, $t3_max_x, $t3_infinity_x);
    }
  } else {
    if (($max_x < $y_min_x) || ($y_max_x < $min_x)) {
      # Empty curve
      return;
    }
    my $x0 = $y_min_x;
    my $x1 = $y_max_x;
    $x0 = $min_x if ($x0 < $min_x);
    $x1 = $max_x if ($max_x < $x1);
    push @intervals,
         clip_t3($x0, $x1, $t3_min_x, $t3_max_x, $t3_infinity_x);
  }
  #print "$#intervals\n";
  for my $i (0 .. ($#intervals-1)/2) {
    my $x_start = $intervals[$i*2];
    my $x_end = $intervals[$i*2+1];
    my $delta = 1.0/$density[$i];
    my $num = int(($x_end-$x_start)/$delta);
    my @surface = ();
    for my $j (0 .. $num) {
      my $t1 = $x_start + $delta * $j;
      my $t2 = ($a * $t1 + $b)/($c * $t1 + $d);
      my $x0 = $s1s[0] + ($s1t[0]-$s1s[0])*$t1;
      my $y0 = $s1s[1] + ($s1t[1]-$s1s[1])*$t1;
      my $z0 = $s1s[2] + ($s1t[2]-$s1s[2])*$t1;
      my $x1 = $s2s[0] + ($s2t[0]-$s2s[0])*$t2;
      my $y1 = $s2s[1] + ($s2t[1]-$s2s[1])*$t2;
      my $z1 = $s2s[2] + ($s2t[2]-$s2s[2])*$t2;
      push @surface, [$x0, $y0, $z0, $x1, $y1, $z1];
    }
    # Last one:
    my $t1 = $x_end;
    my $t2 = ($a * $t1 + $b)/($c * $t1 + $d);
    my $x0 = $s1s[0] + ($s1t[0]-$s1s[0])*$t1;
    my $y0 = $s1s[1] + ($s1t[1]-$s1s[1])*$t1;
    my $z0 = $s1s[2] + ($s1t[2]-$s1s[2])*$t1;
    my $x1 = $s2s[0] + ($s2t[0]-$s2s[0])*$t2;
    my $y1 = $s2s[1] + ($s2t[1]-$s2s[1])*$t2;
    my $z1 = $s2s[2] + ($s2t[2]-$s2s[2])*$t2;
    push @surface, [$x0, $y0, $z0, $x1, $y1, $z1];
    push @surfaces, [@surface];
  }
  return @surfaces;
}

my @surfaces = ();
my $id = 0;

my $lts_res = capture($^X, "lts.pl", "(@s1s,@s1t) (@s2s,@s2t) (@p0,@p1)");
push @surfaces, get_curves($id++, $lts_res);

print '#VRML V2.0 utf8',"\n";
if (!$strict_vrml) {
  print 'DEF CONFIG Configuration {
  accumulation Accumulation { enabled TRUE }
}
DEF BACKGROUND ColorBackground {
  color 1 1 1 0
  clearStencil TRUE
}
Viewpoint {
  type "ORTHOGONAL"
  fieldOfView 3.103
}
DEF SNAP_KEY SingleKeySensor { key "S" }
DEF SNAP Snapshot {
  image Image { }
  fileFormat "png"
  sequence FALSE
  dirName "."
  fileName "3_skew_lines_hos"
}',"\n";
}
print 'DEF EXAMINE_NAV NavigationInfo { type [ "EXAMINE" "ANY" ] }
DEF TRANSFORMER_NAV NavigationInfo { type [ "TRANSFORM" ] }
Transform {
  rotation 1 0 0 -1.30899 # -75
  children [
    Transform {
      rotation 0 0 1 -.261799 # -15
      children [',"\n";
# S1
print '        Shape {
          appearance Appearance {
            material Material { diffuseColor 1 0 0 }
          }
          geometry Extrusion {
	    crossSectionRadius 0.05
	    spine [',"\n";
print "        $s1s[0] $s1s[1] $s1s[2], $s1t[0] $s1t[1] $s1t[2]\n";
print '                  ]
	  }
        }',"\n";
# S2
print '        Shape {
          appearance Appearance {
            material Material { diffuseColor 0 1 0 }
          }
          geometry Extrusion {
	    crossSectionRadius 0.05
	    spine [',"\n";
print "        $s2s[0] $s2s[1] $s2s[2], $s2t[0] $s2t[1] $s2t[2]\n";
print '                  ]
	  }
        }',"\n";
# S3
print '        Shape {
          appearance Appearance {
            material Material { diffuseColor 0 0 1 }
          }
          geometry Extrusion {
	    crossSectionRadius 0.05
	    spine [',"\n";
print "        $p0[0] $p0[1] $p0[2], $p1[0] $p1[1] $p1[2]\n";
print '                  ]
	  }
        }',"\n";
# The surface patches 
print '        Shape {
          appearance DEF RULE_APP Appearance {
            material Material { diffuseColor 0 0 0 }
          }
          geometry IndexedLineSet {
            elliminateHiden TRUE
            coord Coordinate {
              point [',"\n";
# line coordinates
for my $i ( 0 .. $#surfaces ) {
  my $surface_ref = $surfaces[$i];
  for my $j ( 0 .. $#{$surface_ref} ) {
    my $x0 = $surface_ref->[$j][0];
    my $y0 = $surface_ref->[$j][1];
    my $z0 = $surface_ref->[$j][2];
    my $x1 = $surface_ref->[$j][3];
    my $y1 = $surface_ref->[$j][4];
    my $z1 = $surface_ref->[$j][5];
    my $min_t = ($min_z - $z0)/($z1 - $z0);
    my $max_t = ($max_z - $z0)/($z1 - $z0);

    my $min_x = $x0 + ($x1 - $x0) * $min_t;
    my $min_y = $y0 + ($y1 - $y0) * $min_t;

    my $max_x = $x0 + ($x1 - $x0) * $max_t;
    my $max_y = $y0 + ($y1 - $y0) * $max_t;
    print "$min_x $min_y $min_z, $max_x $max_y $max_z\n"
  }
}
print '                    ]
            }
            coordIndex [',"\n";
# line indices
my $k = 0;
for my $i ( 0 .. $#surfaces ) {
  my $surface_ref = $surfaces[$i];
  for my $j ( 0 .. $#{$surface_ref} ) {
    my $u = 2 * $k++;
    my $v = $u+1;
    print "$u $v,\n";
  }
}
print '                       ]
          }
        }',"\n";
# The underlying surface
print '         # Surface
        Shape {
          colorMask 0 0 0 0
          priority 10
          appearance DEF SURFACE_APP Appearance {
            material Material {
              diffuseColor 0.7 0.5 0.6
              transparency 0.4
            }
          }
          geometry DEF GEOMETRY_APP IndexedFaceSet {
            solid FALSE
            polygonOffsetFactor 8
            creaseAngle 2.1
            coord Coordinate {
              point [',"\n";
my $delta_angle = pi * 2 / $surface_density;
my $v_density = $surface_density/2;
my $delta_v = ($v_max - $v_min) / $v_density;
for (my $i = 0; $i < $surface_density; ++$i) {
  my $u = $delta_angle * $i;
  for (my $j = 0; $j < $v_density; ++$j) {
    my $v = $v_min + $delta_v * $j;
    my $x = $radius * (cos($u) - $v * sin($u));
    my $y = $radius * (sin($u) + $v * cos($u));
    my $z = $height * $v;
    print "                     $x $y $z,\n";
  }
  my $v = $v_max;
  my $x = $radius * (cos($u) - $v * sin($u));
  my $y = $radius * (sin($u) + $v * cos($u));
  my $z = $height * $v;
  print "                     $x $y $z,\n";
}
print '                    ]
            }
            coordIndex [',"\n";
for (my $i = 0; $i < $surface_density; ++$i) {
  for (my $j = 0; $j < $v_density; ++$j) {
    my $v0 = $i * ($v_density + 1) + $j;
    my $v1 = $v0 + 1;
    my $v2 = (($i + 1) % $surface_density) * ($v_density + 1) + $j;
    my $v3 = $v2 + 1;
#    print "                        $v0 $v3 $v1 -1, $v0 $v2 $v3 -1\n";
    print "                        $v0 $v1 $v3 $v2 -1\n";
  }
}
print '                       ]
          }
        }
        Shape {
          priority 9
          depthFunction "equal"
          appearance USE SURFACE_APP
          geometry USE GEOMETRY_APP
        }',"\n";
# The axes
print '         # The axes:
        DEF AXES_SWITCH Switch {
          whichChoice -1
          children [
            DEF AXES Transform {
              scale 0.23 0.23 0.23
              children [
                Transform {
                  translation 3.5 0 0
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation 0 0 0
                      children [
                        Shape {
                          appearance DEF TEXT_APP Appearance {
                            material Material { diffuseColor 0 0 0 }
                          }
                          geometry Text {
                            string "X"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  translation 0 3.5 0.4
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation 0 0 -0.2
                      children [
                        Shape {
                          appearance USE TEXT_APP
                          geometry Text {
                            string "Y"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  translation 0 0 3.5
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation -0.1 -0.1 0
                      children [
                        Shape {
                          appearance DEF TEXT_APP Appearance {
                            material Material { diffuseColor 0 0 0 }
                          }
                          geometry Text {
                            string "Z"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Shape {
                  priority 100
                  appearance DEF AXES_APP Appearance {
                    material Material { diffuseColor 0.5 0.5 0.5 }
                  }
                  geometry Sphere { radius 0.1 }
                }
                DEF CS_AXIS Transform {
                  translation 0 1.5 0
                  children [
                    Shape {
                      priority 100
                      appearance USE AXES_APP
                      geometry Cylinder {
                        radius 0.1
                        height 3
                        set_is_bottom_visible FALSE
                        set_is_top_visible FALSE
                      }
                    }
                    Transform {
                      translation 0 1.7 0
                      children [
                        Shape {
                          priority 100
                          appearance USE AXES_APP
                          geometry Cone {
                            bottomRadius 0.2
                            height 0.4
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  rotation 0 0 -1 1.57
                  children [ USE CS_AXIS ]
                }
                Transform {
                  rotation 1 0 0 1.57
                  children [ USE CS_AXIS ]
                }
               ]
            }
           ]
        }
      ]
    }
   ]
}',"\n";
if (!$strict_vrml) {
  print 'ROUTE SNAP_KEY.state TO SNAP.trigger',"\n";
}

exit(0);
