#!/usr/bin/perl -w

use strict;
use warnings;
use IPC::System::Simple qw(system capture);
use Cwd;
use Getopt::Long;
use List::Util qw(max min);
use Math::Complex;
# use Math::Cephes qw(:hypers);

my $help = 0;
my $translate = 0;
my $scale = 1;
$Getopt::Long::ignorecase = 0;
GetOptions("help" => \$help,
           "scale=f" => \$scale,
	   "translate=f" => \$translate) or die "Bad command line option!";
usage() if $help;

my $min_x = 0;
my $max_x = 1;

my $min_y = 0;
my $max_y = 1;

$min_x = ($min_x + $translate) * $scale;
$max_x = ($max_x + $translate) * $scale;
$min_y = ($min_y + $translate) * $scale;
$max_y = ($max_y + $translate) * $scale;

sub usage {
  printf "gen_arr options";
  printf "
\t--help\t\t\tprint this help message
";
  exit(0);
}

sub gcf {
  my ($x, $y) = @_;
  ($x, $y) = ($y, $x % $y) while $y;
  return $x;
}

my @colors = ([1, 0, 0],
	      [0, 1, 0],
	      [0, 0, 1],
	      [1, 1, 0],
	      [1, 0, 1],
	      [0, 1, 1],
	      [0.5, 0.5, 1],
	      [0.5, 1, 0.5],
	      [1, 0.5, 0.5],
	      [0.25, 0.75, 1],
	      [0.25, 1, 0.75],
	      [0.75, 0.25, 1],
	      [1, 0.25, 0.75],
	      [0.75, 1, 0.25],
	      [1, 0.75, 0.25],
	      [0.5, 0.7, 0.8],
	      [0.5, 0.8, 0.7],
	      [0.7, 0.5, 0.8],
	      [0.8, 0.5, 0.7],
	      [0.7, 0.8, 0.5],
	      [0.8, 0.7, 0.5],
	      [0.5, 0.6, 0.9],
	      [0.5, 0.9, 0.6],
	      [0.6, 0.5, 0.9],
	      [0.9, 0.5, 0.6],
	      [0.6, 0.9, 0.5],
	      [0.9, 0.6, 0.5],
	      [0.4, 0.7, 0.9],
	      [0.9, 0.4, 0.7],
	      [0.7, 0.9, 0.4]);

sub clip_t3 {
  my @res = ();
  my ($min_x, $max_x, $t3_min_x, $t3_max_x, $t3_infinity_x) = @_;
  if (($t3_min_x < $t3_infinity_x) && ($t3_infinity_x < $t3_max_x)) {
    # Exclude [$t3_min_x,$t3_max_x]
    if (($t3_min_x < $min_x) && ($max_x < $t3_max_x)) {
      return @res;
    }
    if (($t3_max_x < $min_x) || ($max_x < $t3_min_x)) {
      # print "1 min_x = $min_x, max_x = $max_x\n";
      push @res, $min_x, $max_x;
      return @res;
    }
    if ($min_x < $t3_min_x) {
      # print "2 min_x = $min_x, max_x = $max_x\n";
      push @res, $min_x, $t3_min_x;
    }
    if ($t3_max_x < $max_x) {
      # print "3 min_x = $min_x, max_x = $max_x\n";
      push @res, $t3_max_x, $max_x;
    }
    return @res;
  }
  # Include [$t3_min_x,$t3_max_x]
  $min_x = max($min_x, $t3_min_x);
  $max_x = min($t3_max_x, $max_x);
  # print "4 min_x = $min_x, max_x = $max_x\n";
  if ($min_x < $max_x) {
    push @res, $min_x, $max_x;
  }
  return @res;
}

sub get_curves {
  my $id = $_[0];
  my ($a, $b, $c, $d, $t3_0_x, $t3_1_x, $t3_infinity_x) = split(/ /, $_[1]);
  my $t3_min_x = min($t3_0_x, $t3_1_x);
  my $t3_max_x = max($t3_0_x, $t3_1_x);
  my @intervals = ();

  if ($c == 0) {
    # A horizontal or diagonal line
    if ($d == 0) {
      return "% Illegal curve: Both c and d vanish!";
    }
    my ($y0, $y1);
    if ($a == 0) {
      # A horizontal line
      $y0 = $b / $d;
      $y1 = $y0;
      if (($min_y > $y1) || ($y0 > $max_y)) {
        return "% Empty curve";
      }
      push(@intervals,
	   clip_t3($min_x, $max_x, $t3_min_x, $t3_max_x, $t3_infinity_x));
    } else {
      # A diagonal line
      $a = $a / $d;
      $b = $b / $d;
      my $i0 = -$b / $a;
      my $i1 = (1 - $b) / $a;
      my $x0 = min($i0, $i1);
      my $x1 = max($i0, $i1);
      if (($x0 > 1) || ($x1 < 0)) {
        return "% Empty curve";
      }
      $x0 = $min_x if ($x0 < $min_x);
      $x1 = $max_x if ($max_x < $x1);
      $y0 = $a * $x0 + $b;
      $y1 = $a * $x1 + $b;
      push(@intervals,
	   clip_t3($x0, $x1, $t3_min_x, $t3_max_x, $t3_infinity_x));
    }
    my @result = ();
    for my $i (0 .. $#intervals/2-1) {
      my $x0 = $intervals[$i*2];
      my $x1 = $intervals[$i*2+1];
      my $curve = sprintf("\\pcline(%f, %f)(%f, %f)", $x0, $y0, $x1, $y1);
      push @result, $curve;
    }
    return "@result";
  }

  if ($a * $d == $b * $c) {
    # A vertical line
    my $num = gcf($a, $b);
    my $den = gcf($c, $d);
    my $val = $num / $den;
    if (($min_y > $val) || ($val > $max_y)) {
      return "% Empty curve";
    }
    my $result = sprintf("\\pcline(%f, %f)(%f, %f)", $val, $min_y, $val, $max_y);
    return $result;
  }
  # Hyperbola
  my $y0_x = ($min_x * $d - $b) / ($a - $min_x * $c);
  my $y1_x = ($max_x * $d - $b) / ($a - $max_x * $c);
  my $y_min_x = min($y0_x, $y1_x);
  my $y_max_x = max($y0_x, $y1_x);
  my $asymptote_x = -$d / $c;
  my $asymptote_y = $a / $c;
  # print "y_min_x = $y_min_x\n";
  # print "y_max_x = $y_max_x\n";
  if (($min_y < $asymptote_y) && ($asymptote_y < $max_y)) {
    #  The hyperbola horizontal asymptote is in [min_y,max_y]
    if ($min_x < $y_min_x) {
      my $x1 = $y_min_x;
      $x1 = $max_x if ($max_x < $x1);
      push @intervals,
           clip_t3($min_x, $x1, $t3_min_x, $t3_max_x, $t3_infinity_x);
    }
    if ($y_max_x < $max_x) {
      my $x0 = $y_max_x;
      $x0 = $min_x if ($x0 < $min_x);
      push @intervals,
	   clip_t3($x0, $max_x, $t3_min_x, $t3_max_x, $t3_infinity_x);
    }
  } else {
    if (($max_x < $y_min_x) || ($y_max_x < $min_x)) {
      return "% Empty curve";
    }
    my $x0 = $y_min_x;
    my $x1 = $y_max_x;
    $x0 = $min_x if ($x0 < $min_x);
    $x1 = $max_x if ($max_x < $x1);
    push @intervals,
         clip_t3($x0, $x1, $t3_min_x, $t3_max_x, $t3_infinity_x);
  }
  my @result = ();
  #print "$#intervals\n";
  for my $i (0 .. ($#intervals-1)/2) {
    my $x0 = $intervals[$i*2];
    my $x1 = $intervals[$i*2+1];
    my $cl = "color$id";
    my $curve = sprintf("\\psplot[linecolor=$cl]{%f}{%f}{x $a mul $b add $c x mul $d add div}", $x0, $x1);
    push @result, $curve;
  }
  # print "@result\n";
  return "@result";
}

my @curves;
my $i = 0;
my $id = 0;

# One Connected Component:
# my @s1s = (-3300,1501,0);
# my @s1t = (3300,1502,0);
# my @s2s = (0,-1503,-3300);
# my @s2t = (0,-1504,3300);

# 3 Connected Components:
my @s1s = ( 1,  4, -4);
my @s1t = ( 1, -4,  4);
my @s2s = (-1, -4, -4);
my @s2t = (-1,  4,  4);

my @p0 = (-4,  1, -4);
my @p1 = ( 4,  1,  4);
my @p2 = ( 4, -1, -4);
my @p3 = (-4, -1,  4);

my $lts_res;

# 1: (0,1)
$lts_res = capture($^X, "lts.pl", "(@s1s,@s1t) (@s2s,@s2t) (@p0,@p1)");
push @curves, get_curves($id++, $lts_res);

# 2: (2,3)
#$lts_res = capture($^X, "lts.pl", "(@s1s,@s1t) (@s2s,@s2t) (@p2,@p3)");
#push @curves, get_curves($id++, $lts_res);

my $latex_header = "\\documentclass{article}
\\usepackage{txfonts}
\\usepackage{pstricks}
\\usepackage{color}
\\usepackage{pst-plot}
\\usepackage{pst-tree}
\\usepackage{pst-eps}
\\usepackage{multido}
\\usepackage{pst-node}
\\pagestyle{empty}
% \\newpsstyle{SimpleHandle}{fillstyle=solid,fillcolor=white,framearc=0.5}";

print "$latex_header\n";

for $i ( 0 ..  $#colors ) {
  my $color = $colors[$i];
  my ($r, $g, $b) = @$color;
  my $sum = $r+ $g + $b;
  $r = $r / $sum;
  $g = $g / $sum;
  $b = $b / $sum;
  printf("\\definecolor{color$i}{rgb}{%f,%f,%f}\n", $r,$g,$b);
}
print "\\begin{document}\n";
printf("\\psset{unit=%fcm}\n", 8/$scale);
printf("\\begin{pspicture}(%f,%f)(%f,%f)\n", $min_x, $min_y, $max_x, $max_y);
printf("  \\psframe[linewidth=0.25pt](%f,%f)(%f,%f)", $min_x, $min_y, $max_x, $max_y);

$i = 0;
for my $curve (@curves) {
  print "  $curve % ($i)\n";
  ++$i;
}
printf(" \\psaxes*[linecolor=black,linewidth=0.25pt,labels=none]{->}(0,0)(%f,%f)(%f,%f)\n", $min_x, $min_y, $max_x, $max_y);
print "  \\end{pspicture}\n\\end{document}\n";

exit(0);
