#!/usr/bin/perl

use strict;
use Math::Trig;
use Getopt::Long;

my $help = 0;
my $strict_vrml = 0;

my $radius = 1;
my $height = 1;
my $twist = 1;
my $bounding_box = 0;
my $density = 16;
my $surface_density = 64;

$Getopt::Long::ignorecase = 0;
GetOptions("help" => \$help,
	   "strict-vrml" => \$strict_vrml,
           "bounding-box" => \$bounding_box,
           "radius=f" => \$radius,
           "height=f" => \$height,
           "twist=f" => \$twist,
           "density=i" => \$density) or die "Bad command line option!";
usage() if $help;

sub usage {
  printf "gen_hyperboloid_of_one_sheet options\n";
  printf "  x(u,v) = a(cos u +- v sin u)\n";
  printf "  y(u,v) = a(sin u +- v cos u)\n";
  printf "  z(u,v) = +- c v\n";
  printf "
\t--help\t\t\tprint this help message
\t--strict-vrml\t\tgenerate standard VRML
\t--bounding-box\t\tgenerate the bounding box
\t--v <width>\t\tset the v parameter to <width>
\t--a <depth>\t\tset the a parameter to <depth>
\t--density <density>\tset the density to <density>
";
  exit(0);
}

my $v_min = -$twist;
my $v_max = $twist;

my $x_min = -1;
my $x_max = 1;
my $y_min = -1;
my $y_max = 1;
my $z_min = $height * $v_min;
my $z_max = $height * $v_max;

print '#VRML V2.0 utf8',"\n";
if (!$strict_vrml) {
  print 'DEF CONFIG Configuration {
  accumulation Accumulation { enabled TRUE }
}
DEF BACKGROUND ColorBackground {
  color 1 1 1 0
  clearStencil TRUE
}
Viewpoint {
  type "ORTHOGONAL"
  fieldOfView 3.004
}
DEF SNAP_KEY SingleKeySensor { key "S" }
DEF SNAP Snapshot {
  image Image { }
  fileFormat "jpg"
  sequence FALSE
  dirName "."
  fileName "hyperboloid_of_one_sheet"
}',"\n";
}
print 'DEF EXAMINE_NAV NavigationInfo { type [ "EXAMINE" "ANY" ] }
DEF TRANSFORMER_NAV NavigationInfo { type [ "TRANSFORM" ] }
DirectionalLight {
  direction 0 3 -2
  intensity 0.7
}
DirectionalLight {
  direction 1 1 1
  intensity 0.08
}
DirectionalLight {
  direction -1 -1 -1
  intensity 0.08
}
Transform {
  rotation 1 0 0 -1.30899 # -75
  children [
    Transform {
      rotation 0 0 1 -.261799 # -15
      children [',"\n";
if ($bounding_box) {
  print '        # The bounding box
        Shape {
          appearance Appearance {
            material Material { diffuseColor 0 0 0 }
          }
          geometry IndexedLineSet {
            coord Coordinate {
              point [',"\n";
  print "                     $x_min $y_min $z_min, $x_min $y_min $z_max,\n";
  print "                     $x_min $y_max $z_min, $x_min $y_max $z_max,\n";
  print "                     $x_max $y_min $z_min, $x_max $y_min $z_max,\n";
  print "                     $x_max $y_max $z_min, $x_max $y_max $z_max\n";
  print '                    ]
            }
            coordIndex [0 1, 0 2, 0 4, 1 3, 1 5, 2 3,
                        2 6, 3 7, 4 5, 4 6, 5 7, 6 7]
          }
        }',"\n";
}
print '        # The hyperbolic paraboloid',"\n";
print '        # 1st rules
        Shape {
          priority 8
          appearance DEF RULE_APP Appearance {
            material Material { diffuseColor 0 0 0 }
          }
          geometry IndexedLineSet {',"\n";
if (!$strict_vrml) {
  print '            elliminateHiden TRUE',"\n";
}
print '            coord Coordinate {
              point [',"\n";
my $delta_angle = pi * 2 / $density;
for (my $i = 0; $i < $density; ++$i) {
  my $u = $delta_angle * $i;
  my $x_src = $radius * (cos($u) + $v_min * sin($u));
  my $y_src = $radius * (sin($u) - $v_min * cos($u));
  my $z_src = $height * $v_min;
  my $x_trg = $radius * (cos($u) + $v_max * sin($u));
  my $y_trg = $radius * (sin($u) - $v_max * cos($u));
  my $z_trg = $height * $v_max;
  print "                     $x_src $y_src $z_src, $x_trg $y_trg $z_trg,\n";
}
print '                    ]
            }
            coordIndex [',"\n";
for (my $i = 0; $i < $density; ++$i) {
  my $source_id = 2 * $i;
  my $target_id = $source_id + 1;
  print "                        $source_id $target_id\n";
}
print '                       ]
          }
        }',"\n";
print '        # 2nd rules
          Switch {
            whichChoice 0
            children [
              Shape {
              priority 8
              appearance DEF RULE_APP Appearance {
                material Material { diffuseColor 0 0 0 }
              }
              geometry IndexedLineSet {
                elliminateHiden TRUE
                coord Coordinate {
                  point [',"\n";
my $delta_angle = pi * 2 / $density;
for (my $i = 0; $i < $density; ++$i) {
  my $u = $delta_angle * $i;
  my $x_src = $radius * (cos($u) - $v_min * sin($u));
  my $y_src = $radius * (sin($u) + $v_min * cos($u));
  my $z_src = $height * $v_min;
  my $x_trg = $radius * (cos($u) - $v_max * sin($u));
  my $y_trg = $radius * (sin($u) + $v_max * cos($u));
  my $z_trg = $height * $v_max;
  print "                         $x_src $y_src $z_src, $x_trg $y_trg $z_trg,\n";
}
print '                        ]
                }
                coordIndex [',"\n";
for (my $i = 0; $i < $density; ++$i) {
  my $source_id = 2 * $i;
  my $target_id = $source_id + 1;
  print "                            $source_id $target_id\n";
}
print '                           ]
              }
            }
           ]
        }',"\n";
print '         # Surface
        Shape {',"\n";
if (!$strict_vrml) {
print '          colorMask 0 0 0 0
          priority 10',"\n";
}
print '          appearance DEF SURFACE_APP Appearance {
            material Material {
              diffuseColor 0.7 0.5 0.6
              transparency 0.2
            }
          }
          geometry DEF GEOMETRY_APP IndexedFaceSet {
            solid FALSE
            polygonOffsetFactor 8
            creaseAngle 2.1
            coord Coordinate {
              point [',"\n";
my $delta_angle = pi * 2 / $surface_density;
my $v_density = $surface_density/2;
my $delta_v = ($v_max - $v_min) / $v_density;
for (my $i = 0; $i < $surface_density; ++$i) {
  my $u = $delta_angle * $i;
  for (my $j = 0; $j < $v_density; ++$j) {
    my $v = $v_min + $delta_v * $j;
    my $x = $radius * (cos($u) - $v * sin($u));
    my $y = $radius * (sin($u) + $v * cos($u));
    my $z = $height * $v;
    print "                     $x $y $z,\n";
  }
  my $v = $v_max;
  my $x = $radius * (cos($u) - $v * sin($u));
  my $y = $radius * (sin($u) + $v * cos($u));
  my $z = $height * $v;
  print "                     $x $y $z,\n";
}
print '                    ]
            }
            coordIndex [',"\n";
for (my $i = 0; $i < $surface_density; ++$i) {
  for (my $j = 0; $j < $v_density; ++$j) {
    my $v0 = $i * ($v_density + 1) + $j;
    my $v1 = $v0 + 1;
    my $v2 = (($i + 1) % $surface_density) * ($v_density + 1) + $j;
    my $v3 = $v2 + 1;
#    print "                        $v0 $v3 $v1 -1, $v0 $v2 $v3 -1\n";
    print "                        $v0 $v1 $v3 $v2 -1\n";
  }
}
print '                       ]
          }
        }
        Shape {
          priority 9
          depthFunction "equal"
          appearance USE SURFACE_APP
          geometry USE GEOMETRY_APP
        }',"\n";
print '        # The axes:
        DEF AXES_SWITCH Switch {
          whichChoice 0
          children [
            DEF AXES Transform {
              scale 0.23 0.23 0.23
              children [
                Transform {
                  translation 3.5 0 0
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation 0 0 0
                      children [
                        Shape {
                          priority 100
                          appearance DEF TEXT_APP Appearance {
                            material Material { diffuseColor 0 0 0 }
                          }
                          geometry Text {
                            string "X"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  translation 0 3.5 0.4
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation 0 0 -0.2
                      children [
                        Shape {
                          priority 100
                          appearance USE TEXT_APP
                          geometry Text {
                            string "Y"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  translation 0 0 3.5
                  rotation 0 0 1 .261799 # 15
                  children [
                    Transform {
                      rotation 1 0 0 1.30899 # 75
                      translation -0.1 -0.1 0
                      children [
                        Shape {
                          priority 100
                          appearance DEF TEXT_APP Appearance {
                            material Material { diffuseColor 0 0 0 }
                          }
                          geometry Text {
                            string "Z"
                            fontStyle FontStyle { size 0.5 }
                          }
                        }
                       ]
                    }
                   ]
                }
                Shape {
                  priority 100
                  appearance DEF AXES_APP Appearance {
                    material Material { diffuseColor 0.5 0.5 0.5 }
                  }
                  geometry Sphere { radius 0.1 }
                }
                DEF CS_AXIS Transform {
                  translation 0 1.5 0
                  children [
                    Shape {
                      priority 100
                      appearance USE AXES_APP
                      geometry Cylinder {
                        radius 0.1
                        height 3
                        set_is_bottom_visible FALSE
                        set_is_top_visible FALSE
                      }
                    }
                    Transform {
                      translation 0 1.7 0
                      children [
                        Shape {
                          priority 100
                          appearance USE AXES_APP
                          geometry Cone {
                            bottomRadius 0.2
                            height 0.4
                          }
                        }
                       ]
                    }
                   ]
                }
                Transform {
                  rotation 0 0 -1 1.57
                  children [ USE CS_AXIS ]
                }
                Transform {
                  rotation 1 0 0 1.57
                  children [ USE CS_AXIS ]
                }
               ]
            }
           ]
        }
      ]
    }
   ]
}',"\n";
if (!$strict_vrml) {
  print 'ROUTE SNAP_KEY.state TO SNAP.trigger',"\n";
}

exit(0);
